'''
WikiLinks Extension for Python-Markdown
======================================

Converts [[WikiLinks]] and [[label|wikilinks]] to relative links.

Derived from <https://Python-Markdown.github.io/extensions/wikilinks>

Original code Copyright [Waylan Limberg](http://achinghead.com/).

All changes Copyright The Python Markdown Project

License: [BSD](https://opensource.org/licenses/bsd-license.php)

'''

from markdown.extensions import Extension
from markdown.inlinepatterns import InlineProcessor
import xml.etree.ElementTree as etree
import urllib.parse
from . import backend

# action keyword in wikilinks
_ACTION_KW = 'action:'


def build_url(target):
    """ Build a url from the wiki target. """
    if target.startswith(_ACTION_KW):
        target = target[len(_ACTION_KW):]
        if not target:
            return None
        return '../' + urllib.parse.quote(target, safe='/#?=')

    # anchor link in the current page, return the anchor
    if target.startswith('#'):
        return urllib.parse.quote(target, safe='#')

    # default action is to view the page
    return '../view/' + urllib.parse.quote(target, safe='#')


def is_dangling(target):
    '''Return True if the wiki page is missing
    '''
    # assume action link are ok
    if target.startswith(_ACTION_KW):
        return False

    pagename = target.split('#')[0]
    if not pagename or pagename in backend.get_pages_list():
        return False

    return True


class WikiLinkExtension(Extension):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def extendMarkdown(self, md):
        self.md = md

        # append to end of inline patterns
        WIKILINK_RE = r'\[\[\s*((?P<label>[^][]+)\s*\|\s*)?(?P<target>[^][|]+?)\s*\]\]'
        wikilinkPattern = WikiLinksInlineProcessor(WIKILINK_RE)
        wikilinkPattern.md = md
        md.inlinePatterns.register(wikilinkPattern, 'wikilink', 75)


class WikiLinksInlineProcessor(InlineProcessor):
    def __init__(self, pattern):
        super().__init__(pattern)

    def handleMatch(self, m, data):
        mdict = m.groupdict()
        target = mdict['target']
        label = mdict['label'] or target
        url = build_url(target)
        if url:
            a = etree.Element('a')
            a.text = label
            a.set('href', url)
            link_class = ['wikilink']
            if is_dangling(target):
                link_class.append('text-danger')
            a.set('class', ' '.join(link_class))
        else:
            a = ''
        return a, m.start(0), m.end(0)


def makeExtension(**kwargs):  # pragma: no cover
    return WikiLinkExtension(**kwargs)



# A Bottle of Wiki — personal wiki
# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2019-2021  Benoît Monin <benoit.monin@gmx.fr>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import gettext
import locale

from . import alerts
from . import config


def deferred_gettext(message):
    '''Pass-through to be used before gettext is installed
    '''
    return message


def init():
    '''Setup the locale and install the translation
    '''
    # setup the locale
    try:
        lang = locale.setlocale(locale.LC_ALL, config.get('abow.locale'))
    except locale.Error:
        lang = locale.setlocale(locale.LC_ALL, '')
        alerts.add('warning',
                   __('Invalid locale "{cfg}", using "{dflt}"').format(
                       cfg=config.get('abow.locale'), dflt=lang))

    # let the time in C locale for bottle 'last-modified' html header
    locale.setlocale(locale.LC_TIME, 'C')

    # install the translation
    translation = gettext.translation('abow',
                                      localedir=config.get('abow.locale_path'),
                                      languages=[lang],
                                      fallback=True)
    translation.install()


def get_lang():
    '''Return the html lang attributes from the current locale
    '''
    try:
        return locale.getlocale()[0].replace('_', '-')
    except AttributeError:
        return 'en'



# A Bottle of Wiki — personal wiki
# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2019-2021  Benoît Monin <benoit.monin@gmx.fr>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import markdown

from markdown.extensions.abbr import AbbrExtension
from markdown.extensions.admonition import AdmonitionExtension
from markdown.extensions.attr_list import AttrListExtension
from markdown.extensions.def_list import DefListExtension
from markdown.extensions.footnotes import FootnoteExtension
from markdown.extensions.md_in_html import MarkdownInHtmlExtension
from markdown.extensions.sane_lists import SaneListExtension
from markdown.extensions.toc import TocExtension

from pymdownx.betterem import BetterEmExtension
from pymdownx.caret import InsertSupExtension
from pymdownx.escapeall import EscapeAllExtension
from pymdownx.highlight import HighlightExtension
from pymdownx.keys import KeysExtension
from pymdownx.mark import MarkExtension
from pymdownx.superfences import SuperFencesCodeExtension
from pymdownx.tabbed import TabbedExtension
from pymdownx.tasklist import TasklistExtension
from pymdownx.tilde import DeleteSubExtension

from .mdx_tables import TableExtension
from .mdx_wikilinks import WikiLinkExtension


# markdown renderer
_md = None


def init():
    '''Initialize the renderer if needed
    '''
    global _md

    if _md is None:
        _md = markdown.Markdown(
            output_format='html5',
            extensions=[
                AbbrExtension(),
                AdmonitionExtension(),
                AttrListExtension(),
                DefListExtension(),
                FootnoteExtension(BACKLINK_TEXT='&#8617;&#xFE0E;',
                                  BACKLINK_TITLE=_('Jump back to reference %d')),
                MarkdownInHtmlExtension(),
                SaneListExtension(),
                TocExtension(marker='',
                             permalink='#',
                             permalink_class='permalink',
                             permalink_title='',
                             toc_depth=5,
                             baselevel=2),

                BetterEmExtension(),
                InsertSupExtension(),
                EscapeAllExtension(hardbreak=True, nbsp=True),
                HighlightExtension(linenums_style='pymdownx-inline'),
                KeysExtension(),
                MarkExtension(),
                SuperFencesCodeExtension(),
                TabbedExtension(),
                TasklistExtension(custom_checkbox=True),
                DeleteSubExtension(),

                WikiLinkExtension(),
                TableExtension(),
                ],
            )


def render(text):
    '''Render some text to html

    Return a tuple containing the rendered html and the table of content
    '''
    init()
    _md.reset()
    html = _md.convert(text)
    return {
        'html': html,
        'toc': _md.toc,
        }



# A Bottle of Wiki — personal wiki
# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2019-2021  Benoît Monin <benoit.monin@gmx.fr>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import errno
import json
import locale
import time
import uuid

from . import alerts
from . import config

# page content file extension
_CONTENT_EXT = '.md'

# page attributes file extension
_ATTR_EXT = '.json'

# pages attributes dictionary
# the key is the pagename, the value is a dictionary of attributes
_pages_attrs = {}


# time reference for the backend modification time
# use the application start time
_ref_mtime = time.time()


def _build_path(pagename='', ext=''):
    '''Return the path of a page
    '''
    return os.path.join(config.get('abow.backend_path'), pagename + ext)


def safe_write(filepath, content):
    '''Save some text content to a file safely

    Raises OSError if the file cannot be written
    '''
    tmppath = filepath + uuid.uuid4().hex

    with open(tmppath, 'w', encoding='utf8') as f:
        f.write(content)
        f.flush()
        os.fsync(f.fileno())

    os.replace(tmppath, filepath)


def get_content(pagename):
    '''Return the content of a page or an empty string if the page is missing
    '''
    fp = _build_path(pagename, _CONTENT_EXT)
    try:
        with open(fp, 'r', encoding='utf8') as f:
            return f.read()
    except OSError as err:
        if err.errno != errno.ENOENT:
            alerts.add('warning',
                       __('Failed to read the content of page {}: {}.').format(
                           pagename, err.strerror))
        return ''


def set_content(pagename, content):
    '''Save the content of a page, create it if needed
    If the content is an empty string, delete the page
    '''
    fp = _build_path(pagename, _CONTENT_EXT)
    if content == '':
        delete(pagename)
    else:
        try:
            safe_write(fp, content.replace('\r', ''))
        except OSError as err:
            alerts.add('danger',
                       __('Failed to save the content of page {}: {}.').format(
                           pagename, err.strerror))


def delete(pagename):
    '''Remove a page from storage and the attributes dict

    Raises OSError if the page cannot be deleted
    '''
    try:
        os.unlink(_build_path(pagename, _CONTENT_EXT))
    except OSError as err:
        if err.errno != errno.ENOENT:
            alerts.add('warning',
                       __('Failed to delete the content of page {}: {}.').format(
                           pagename, err.strerror))
    try:
        os.unlink(_build_path(pagename, _ATTR_EXT))
    except OSError as err:
        if err.errno != errno.ENOENT:
            alerts.add('warning',
                       __('Failed to delete the attributes of page {}: {}.').format(
                           pagename, err.strerror))
    try:
        del _pages_attrs[pagename]
    except KeyError:
        pass


def _load_attrs(pagename):
    '''Load the page attributes from file if needed
    '''
    if pagename not in _pages_attrs:
        fp = _build_path(pagename, _ATTR_EXT)
        try:
            with open(fp, 'r', encoding='utf8') as f:
                _pages_attrs[pagename] = json.load(f)
        except OSError as err:
            if err.errno != errno.ENOENT:
                alerts.add('warning',
                           __('Failed to load the attributes of page {}: {}.').format(
                               pagename, err.strerror))
            _pages_attrs[pagename] = {}


def get_attr(pagename, attr, default=None):
    '''Return one attribute of a page if found, or the default value
    '''
    _load_attrs(pagename)
    attrs = _pages_attrs.get(pagename, {})
    return attrs.get(attr, default)


def set_attr(pagename, attr, value):
    '''Set one attribute of a page
    '''
    _load_attrs(pagename)

    attrs = _pages_attrs.get(pagename, {})
    attrs[attr] = value
    _pages_attrs[pagename] = attrs

    fp = _build_path(pagename, _ATTR_EXT)
    try:
        safe_write(fp, json.dumps(_pages_attrs[pagename]))
    except OSError as err:
        alerts.add('danger',
                   __('Failed to save the attributes of page {}: {}.').format(
                       pagename, err.strerror))


def _list_files():
    '''Return the list of files in the backend

    Note: the list is not sorted
    '''
    # empty pagename and extension gives the containing folder
    top = _build_path()
    try:
        _, _, filenames = next(os.walk(top))
    except StopIteration:
        filenames = []
    return filenames


def get_pages_list():
    '''Return the list of pages

    Note: the list is sorted
    '''
    # empty pagename and extension gives the containing folder
    filenames = _list_files()
    pages = [os.path.splitext(f)[0] for f in filenames if f.endswith(_CONTENT_EXT)]
    return sorted(pages, key=locale.strxfrm)


def get_mtime(pagename=None):
    '''Return the page modification time of a page or the whole backend

    Note: the mtime is the unix epoch as an integer
    '''
    mtime = [_ref_mtime]

    if pagename and os.path.exists(_build_path(pagename, _CONTENT_EXT)):
        paths = [
            _build_path(pagename, _ATTR_EXT),
            _build_path(pagename, _CONTENT_EXT),
            ]
    else:
        paths = [_build_path(f) for f in _list_files()]

    for fp in paths:
        try:
            mtime.append(int(os.path.getmtime(fp)))
        except OSError:
            pass

    return max(mtime)


def search_pages(text):
    '''search pages containing some text

    Note: the list is sorted
    '''
    matching_pages = []
    all_pages = get_pages_list()
    text = text.casefold()

    for page in all_pages:
        content = get_content(page).casefold()
        if text in content:
            matching_pages.append(page)

    return matching_pages



# A Bottle of Wiki — personal wiki
# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2019-2021  Benoît Monin <benoit.monin@gmx.fr>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import setuptools

here = os.path.abspath(os.path.dirname(__file__))

# Get the long description from the README file
with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setuptools.setup(
    name='abow',
    version='0.4.1',
    description='A Bottle of Wiki — personal wiki',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://pypi.org/project/abow/',
    author='Benoît Monin',
    author_email='benoit.monin@gmx.fr',
    license='AGPL-3.0-or-later',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: End Users/Desktop',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content :: Wiki',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Application',
        'Framework :: Bottle',
        'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    keywords='wiki bottle markdown',
    packages=setuptools.find_packages(),
    include_package_data=True,
    python_requires='>=3.5',
    install_requires=[
        'bottle >= 0.12.0',
        'importlib_metadata ~= 1.0; python_version < "3.8"',
        'importlib_resources ~= 1.0; python_version < "3.7"',
        'Markdown >= 3.2',
        'pymdown-extensions >= 7.0',
    ],
    extras_require={
        'dev': [
            'coverage',
            'flake8',
            'pytest',
            ],
        'extra': ['Pygments >= 2.4'],
    },
)



# A Bottle of Wiki — personal wiki
# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2019-2021  Benoît Monin <benoit.monin@gmx.fr>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import collections

# the queue of alerts
_queue = collections.deque()


def add(lvl, txt):
    '''Add a new alert to the queue
    '''
    global _queue
    _queue.append((lvl, txt))


def get_list():
    '''Return the list of alerts
    '''
    global _queue
    return list(_queue)


def clear():
    '''Clear the queue of alerts
    '''
    global _queue
    _queue.clear()



import time
import abow


def test_empty(backend_path):
    '''Test with an empty backend
    '''
    assert abow.backend.get_pages_list() == []
    assert abow.backend.get_mtime() == abow.backend._ref_mtime
    assert abow.backend.get_content('page1') == ''


def test_set_content(backend_path):
    '''Create some content and read it back
    '''
    abow.backend.set_content('page1', 'abc')
    assert abow.backend.get_content('page1') == 'abc'
    abow.backend.set_content('page1', 'def')
    assert abow.backend.get_content('page1') == 'def'
    abow.backend.set_content('page1', '')
    assert 'page1' not in abow.backend.get_pages_list()


def test_delete(backend_path):
    '''Test page deletion
    '''
    abow.backend.set_content('page1', 'abc')
    abow.backend.delete('page1')
    assert abow.backend.get_content('page1') == ''

    # make sure no exception is raised when deleting a non existing page
    abow.backend.delete('page1')


def test_set_attrs(backend_path):
    ''' test page attributes
    '''
    abow.backend.set_attr('page1', 'attr1', 'abc')
    assert abow.backend.get_attr('page1', 'attr1') == 'abc'

    # force a read from backend
    abow.backend._pages_attrs.clear()
    assert abow.backend.get_attr('page1', 'attr1') == 'abc'

    abow.backend.delete('page1')
    assert abow.backend.get_attr('page1', 'attr1') is None


def test_mtime(backend_path):
    ''' test backend mtime
    '''
    cur_time = int(time.time())
    abow.backend.set_content('page1', 'abc')
    assert abow.backend.get_mtime() >= cur_time
    assert abow.backend.get_mtime('page1') >= cur_time


def test_search(backend_path):
    '''test page search
    '''
    abow.backend.set_content('page1', 'abc')
    abow.backend.set_content('page2', 'abcd')

    assert abow.backend.search_pages('abc') == ['page1', 'page2']
    assert abow.backend.search_pages('d') == ['page2']
    assert abow.backend.search_pages('e') == []


def test_set_content_error(invalid_backend_path):
    '''check that an alert is raised if the content cannot be written or read
    '''
    abow.alerts.clear()
    abow.backend.set_content('page1', 'abc')
    assert len(abow.alerts.get_list()) == 1
    assert abow.backend.get_content('page1') == ''
    assert len(abow.alerts.get_list()) == 2


def test_set_attr_error(invalid_backend_path):
    '''check alerts when trying to access attribute
    '''
    abow.alerts.clear()
    # force a read from backend
    abow.backend._pages_attrs.clear()
    assert abow.backend.get_attr('page1', 'attr1') is None
    abow.backend.set_attr('page1', 'attr1', 'abc')
    assert len(abow.alerts.get_list()) == 2


def test_delete_error(invalid_backend_path):
    '''check that an alert is raised if the page cannot be deleted
    '''
    abow.alerts.clear()
    abow.backend.delete('page1')
    assert len(abow.alerts.get_list()) == 2


def test_list_error(invalid_backend_path):
    '''test pages list when the backend is invalid
    '''
    assert abow.backend.get_pages_list() == []
    assert abow.backend.get_mtime() == abow.backend._ref_mtime
